from __future__ import (absolute_import)
from . import analysis
from . import em_framework
from .em_framework import (Model, RealParameter, CategoricalParameter, 
                           IntegerParameter, perform_experiments, optimize,
                           ScalarOutcome, TimeSeriesOutcome, Constant, 
                           Scenario, Policy, MultiprocessingEvaluator, 
                           IpyparallelEvaluator)

from . import util
from .util import save_results, load_results, ema_logging, EMAError

__version__ = '0.9.4'